package meerkat.mixer.proofs;

import meerkat.crypto.concrete.ECElGamalEncryption;
import meerkat.crypto.concrete.Util;
import meerkat.protobuf.ConcreteCrypto.GroupElement;
import meerkat.protobuf.Mixing;
import meerkat.protobuf.Mixing.Mix2Proof.AndProof;
import meerkat.protobuf.Mixing.Mix2Proof.DlogProof;
import org.bouncycastle.math.ec.ECPoint;
import org.factcenter.qilin.primitives.concrete.ECGroup;

import java.math.BigInteger;
import java.util.Random;

import static meerkat.protobuf.Mixing.Mix2Proof.DlogProof.*;

/**
 * Prover, Simulator and Verifier
 */
public class ECElGamalMixProtocols {
    final ECElGamalMixParams params;
    final ECGroup group;
    final ECElGamalEncryption encryptor;
    final ECPoint g;
    final ECPoint h;
    final Random rand;


    final ChallengeGenerator challengeGenerator = new ChallengeGenerator();

    public ECElGamalMixProtocols(ECElGamalMixParams params, Random rand) {
        this.params = params;
        this.rand = rand;
        group = params.group;
        encryptor = params.encryptor;
        g = params.g;
        h = params.h;
    }

    public class ChallengeGenerator implements SigmaProtocolOr2.ChallengeGenerator {
        @Override
        public BigInteger generateChallenge() { return encryptor.generateRandomExponent(rand); }

        @Override
        public BigInteger subtractChallenge(BigInteger c1, BigInteger c2) { return c1.subtract(c2).mod(group.orderUpperBound()); }
    }


    public class DlogStatementSchnorrProver implements SigmaProtocol.Prover<DlogProof.FirstMessage, DlogProof.FinalMessage> {
        ECElGamalMixParams.DlogStatement statement;
        ECElGamalMixParams.DlogStatementWitness witness;

        BigInteger r = null;

        public DlogStatementSchnorrProver(ECElGamalMixParams.DlogStatement statement, ECElGamalMixParams.DlogStatementWitness witness) {
            this.statement = statement;
            this.witness = witness;
        }

        @Override
        public FirstMessage getFirstMessage() {
            r = encryptor.generateRandomExponent(rand);
            ECPoint gr = group.multiply(statement.g, r);
            ECPoint hr = group.multiply(statement.h, r);

            FirstMessage firstMessage = FirstMessage.newBuilder()
                    .setGr(encryptor.encodeElement(gr))
                    .setHr(encryptor.encodeElement(hr))
                    .build();

            return firstMessage;
        }

        @Override
        public FinalMessage getFinalMessage(BigInteger challenge) {
            return FinalMessage.newBuilder()
                .setXcr(Util.encodeBigInteger(challenge.multiply(witness.x).add(r).mod(group.orderUpperBound())))
                .build();
        }

        @Override
        public void reset() { r = null; }
    }

    public class DlogStatementSchnorrVerifier implements SigmaProtocol.Verifier<DlogProof.FirstMessage, DlogProof.FinalMessage>  {
        final ECElGamalMixParams.DlogStatement statement;

        public DlogStatementSchnorrVerifier(ECElGamalMixParams.DlogStatement statement) {
            this.statement = statement;
        }

        @Override
        public boolean verify(FirstMessage firstMessage, BigInteger challenge,
                              FinalMessage finalMessage) {
            GroupElement grEncoded = firstMessage.getGr();
            ECPoint gr = encryptor.decodeElement(grEncoded);

            GroupElement hrEncoded = firstMessage.getHr();
            ECPoint hr = encryptor.decodeElement(hrEncoded);

            BigInteger xcr = Util.decodeBigInteger(finalMessage.getXcr());

            boolean gGood = group.add(gr, group.multiply(statement.a,challenge)).equals(group.multiply(statement.g,xcr));
            boolean hGood = group.add(hr, group.multiply(statement.b,challenge)).equals(group.multiply(statement.h,xcr));
            return gGood && hGood;
        }

    }

    public class DlogStatementSchnorrSimulator implements SigmaProtocol.Simulator<DlogProof.FirstMessage, DlogProof.FinalMessage> {
        ECElGamalMixParams.DlogStatement statement;
        BigInteger response = null;

        public DlogStatementSchnorrSimulator(ECElGamalMixParams.DlogStatement statement) {
            this.statement = statement;
        }

        @Override
        public FirstMessage getFirstMessage(BigInteger challenge) {
            response = encryptor.generateRandomExponent(rand);

            ECPoint u = group.multiply(statement.g, response).subtract(group.multiply(statement.a,challenge));
            ECPoint v = group.multiply(statement.h, response).subtract(group.multiply(statement.b,challenge));
            return FirstMessage.newBuilder()
                    .setGr(encryptor.encodeElement(u))
                    .setHr(encryptor.encodeElement(v))
                    .build();
        }

        @Override
        public FinalMessage getFinalMessage() {

            return FinalMessage.newBuilder()
                    .setXcr(Util.encodeBigInteger(response))
                    .build();
        }

        @Override
        public void reset() { response = null; }
    }

    /**
     * Prover for plaintext equivalence of a pair of ciphertexts.
     */
    public class AndStatementProver
            extends SigmaProtocolAnd2.Prover<AndProof.FirstMessage, DlogProof.FirstMessage, AndProof.FinalMessage, DlogProof.FinalMessage> {
        public AndStatementProver(ECElGamalMixParams.AndStatement statement,
                                  ECElGamalMixParams.AndStatementWitness witness) {
            super(ProtobufConcatenators.concatAnd1, ProtobufConcatenators.concatAnd2, new DlogStatementSchnorrProver(statement.clauses[0], witness.witnesses[0]),
                    new DlogStatementSchnorrProver(statement.clauses[1], witness.witnesses[1]));
        }
    }

    public class AndStatementVerifier
            extends SigmaProtocolAnd2.Verifier<AndProof.FirstMessage, DlogProof.FirstMessage, AndProof.FinalMessage, DlogProof.FinalMessage> {
        public AndStatementVerifier(ECElGamalMixParams.AndStatement statement) {
            super(ProtobufConcatenators.concatAnd1, ProtobufConcatenators.concatAnd2,
                    new DlogStatementSchnorrVerifier(statement.clauses[0]), new DlogStatementSchnorrVerifier(statement.clauses[1]));
        }
    }

    public class AndStatementSimulator extends SigmaProtocolAnd2.Simulator {
        public AndStatementSimulator(ECElGamalMixParams.AndStatement statement) {
            super(ProtobufConcatenators.concatAnd1, ProtobufConcatenators.concatAnd2,
                    new DlogStatementSchnorrSimulator(statement.clauses[0]), new DlogStatementSchnorrSimulator(statement.clauses[1]));
        }
    }

    public class Mix2Prover extends SigmaProtocolOr2.Prover<Mixing.Mix2Proof.FirstMessage, Mixing.Mix2Proof.AndProof.FirstMessage, Mixing.Mix2Proof.FinalMessage, Mixing.Mix2Proof.AndProof.FinalMessage> {
        public Mix2Prover(ECElGamalMixParams.Mix2Statement statement, ECElGamalMixParams.Mix2StatementWitness witness) {
            super(ProtobufConcatenators.concatMix1, ProtobufConcatenators.concatMix2, ECElGamalMixProtocols.this.challengeGenerator,
                    new AndStatementProver(statement.clauses[witness.trueClauseIndex], witness.witness),
                    new AndStatementSimulator(statement.clauses[1 - witness.trueClauseIndex]),
                    witness.trueClauseIndex);
        }
    }

    public class Mix2Verifier extends SigmaProtocolOr2.Verifier<Mixing.Mix2Proof.FirstMessage, Mixing.Mix2Proof.AndProof.FirstMessage, Mixing.Mix2Proof.FinalMessage, Mixing.Mix2Proof.AndProof.FinalMessage> {
        public Mix2Verifier(ECElGamalMixParams.Mix2Statement statement) {
            super(ProtobufConcatenators.concatMix1, ProtobufConcatenators.concatMix2, ECElGamalMixProtocols.this.challengeGenerator,
                    new AndStatementVerifier(statement.clauses[0]), new AndStatementVerifier(statement.clauses[1]));
        }
    }
}
